// Dependencies
//#region Dependencies

//  External dependencies
var fs = require('fs');
var cp = require('child_process');
var moment = require('moment');
var path = require('path');

// Local dependencies
var Logger;// Used for logging
var SocketIo;// Used for messaging
var NetSnmp;// Used for SNMP messaging - base library
var DbManager;// SqlLite database manager
var SmtpManager;// Used for SMTP messaging
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions

//#endregion Dependencies

// Constants
//#region Constants

//#region Constants - Devices
// List of supported devices
var SerialPortDeviceOptions = [];

//#endregion Constants - Devices

//#region Constants - Serial Port
var SERIAL_PORTS_TOTAL = 8;
var SERIAL_PORT_NO_DEVICE = "NONE";
var SERIAL_PORT_DEVICE_FILE = "./data/port"

// Serial Port settings - used to track setttings on each port
//  - In memory for API access and port amangements
//  - Persisted in summit.db 
var SerialPortSettings = [
    { "Address": "0", "Name": "Port 1", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 },
    { "Address": "1", "Name": "Port 2", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 },
    { "Address": "2", "Name": "Port 3", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 },
    { "Address": "3", "Name": "Port 4", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 },
    { "Address": "4", "Name": "Port 5", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 },
    { "Address": "5", "Name": "Port 6", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 },
    { "Address": "6", "Name": "Port 7", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 },
    { "Address": "7", "Name": "Port 8", "Enabled": "FALSE", "SendEvents": "FALSE", "Device": "NONE", "TimeChanged": 0 }
];

// Serial Port device objects - used to track device drivers on each port
//  - In memory for device management
//  - Persistence is not required
var SerialPortDeviceObjects = [null, null, null, null, null, null, null, null];
var SERIAL_PORT_LOAD_DEFAULT_MODEL = 0;
var SERIAL_PORT_LOAD_PERSISTED_MODEL = 1;

//#endregion Constants - Serial Port

//#endregion Constants

// Constructor
//#region Constructor

// Initialize the analog outputs upon system start
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates analogOutputsArray
exports.InitializeSerialPorts = function (smtp, db, io, logger, netSnmp, callback) {

    // Initialize constants
    //#region Initialize constants
    SmtpManager = smtp;
    DbManager = db;
    SocketIo = io;
    Logger = logger;
    NetSnmp = netSnmp
    //#endregion Initialize constants

    // Load device models
    //#region Load device models
    LoadSerialPortDeviceOptions();
    //#endregion Load device models

    // Read SerialPorts table
    //#region Create/Read SerialPorts persistent data
    DbManager.get("SELECT name FROM sqlite_master WHERE type='table' AND name='SerialPorts'", function (error, rows) {
        // Error reading SerialPorts table
        //#region Error reading SerialPorts table
        if (error !== null) {
            logger.error('serialports.InitializeSerialPorts: Check SerialPorts table - error=' + JSON.stringify(error));
        }
        //#endregion Error reading SerialPorts table

        // Create SerialPorts table, because nothing exists so far ...
        //#region Create SerialPorts table
        else if (rows === undefined) {
            logger.info("serialports.InitializeSerialPorts: Create SerialPorts table ...");
            DbManager.exec("CREATE TABLE if not exists SerialPorts (Address INTEGER, Name TEXT, Enabled TEXT, SendEvents TEXT, Device TEXT, TimeChanged TEXT )", function (error) {
                if (error !== null) {
                    logger.error('serialports.InitializeSerialPorts: Create SerialPorts table - error=' + JSON.stringify(error));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the SerialPort row was changed

                    // Created the table, now add data ...
                    DbManager.run("BEGIN")
                    for (var i = 0; i < SERIAL_PORTS_TOTAL; i++) {
                        // Initialize the database
                        insertStatement = i + ", 'Port " + (i + 1) + "', 'FALSE', 'FALSE', 'NONE', '" + timeChanged + "'";
                        DbManager.run("INSERT INTO SerialPorts (Address, Name, Enabled, SendEvents, Device, TimeChanged) VALUES (" + insertStatement + ")", function (error) {
                            if (error !== null) {
                                logger.error('serialports.InitializeSerialPorts: Insert SerialPorts table - error=' + JSON.stringify(error));
                            }
                        });
                    }
                    DbManager.run("COMMIT");

                    // SerialPorts table created
                    if (typeof callback == "function") {
                        logger.info('serialports.InitializeSerialPorts: Initialized new SerialPorts table');
                        callback();
                    }
                }
            });
        }
        //#endregion Create SerialPorts table

        // Read existing SerialPorts persistent data
        //#region Read existing SerialPorts persistent data
        else {
            logger.info("serialports.InitializeSerialPorts: Read SerialPorts table ...");
            // Read 
            DbManager.each("SELECT rowid as id, Address, Name, Enabled, SendEvents, Device, TimeChanged FROM SerialPorts ORDER BY Address", function (error, row) {
                if (error !== null) {
                    logger.error('serialports.InitializeSerialPorts: Read SerialPorts table - error=' + JSON.stringify(error));
                }
                else {
                    // Load Serial Port data
                    //#region Load Serial Port data
                    var _port = row.Address;
                    var _device = row.Device;
                    SerialPortSettings[_port].Name = row.Name;
                    SerialPortSettings[_port].Enabled = row.Enabled;
                    SerialPortSettings[_port].SendEvents = row.SendEvents;
                    SerialPortSettings[_port].Device = _device;
                    SerialPortSettings[_port].TimeChanged = row.TimeChanged;
                    //#endregion Load Serial Port data

                    // Instantiate Device Object, Load Device Model and Start Operation
                    //#region Instantiate Device Object, Load Device Model and Start Operation
                    if (_device != SERIAL_PORT_NO_DEVICE) {
                        require("./devices/" + _device + ".js")(_port, SERIAL_PORT_LOAD_PERSISTED_MODEL, Logger, SocketIo, SmtpManager, NetSnmp, function (deviceObject) {
                            // Instantiate an object for this device
                            SerialPortDeviceObjects[_port] = deviceObject;// Track the device object

                            // Load the device model
                            var _devicemodel = SerialPortDeviceObjects[_port].GetDeviceModel(); // Load the persistent port data
                            Logger.info("serialports.InitializeSerialPorts: Port[" + _port + "]=" + SerialPortSettings[_port].Device + " (uuid=" + _devicemodel.uuid + ") was instantiated!");

                            // Enable processing/eventing
                            SerialPortDeviceObjects[_port].Enable(SerialPortSettings[_port].Enabled);// Start or stop the device
                            SerialPortDeviceObjects[_port].SendEvents(SerialPortSettings[_port].SendEvents);// SendEvents or not from the device
                            SerialPortDeviceObjects[_port].SetSnmpHostSessionsArray(SnmpHostSessionsArray, function () { });// Update the SNMP Sessions Array
                        });
                    }
                    //#endregion Instantiate Device Object, Load Device Model and Start Operation
                }
            }, function () {
                // Serial Ports persistence read is complete
                //#region Serial Ports persistence read is complete
                if (typeof callback == "function") {
                    logger.info('serialports.InitializeSerialPorts:  Initialized SerialPortSettings from existing SerialPorts table');
                    callback();
                }
                //#endregion Serial Ports persistence read is complete
            });
        }
        //#endregion Read existing SerialPorts persistent data
    });
    //#endregion Create/Read SerialPorts persistent data
}
//#endregion Constructor

//#region Methods

// Device Option methods - load/get device options/drivers
//#region Device Option methods
// Load list of supported devices from the /devices directory
module.exports.LoadSerialPortDeviceOptions = LoadSerialPortDeviceOptions;
function LoadSerialPortDeviceOptions() {

    var serialPortDeviceOptions = [];
    var deviceOptionFound;
    fs.readdir("./devices", function (err, files) {
        var count = 0;
        for (var i = 0; i < files.length; i++) {
            if (path.extname(files[i]) !== '.json') continue;// Only consider JSON files
            // Get the root 
            deviceOptionFound = require("./devices/" + files[i]);

            // Read file and store name in an array
            serialPortDeviceOptions.push(deviceOptionFound);
            Logger.info("serialports.LoadSerialPortDeviceOptions: Loading Device[" + count + "] (File=" + files[i] + "): id=" + deviceOptionFound.id + " (" + deviceOptionFound.title + " v" + deviceOptionFound.version + " - " + deviceOptionFound.date + ")");
            count++;
        }

        SerialPortDeviceOptions = serialPortDeviceOptions;
    });
}

exports.GetSerialPortDeviceOptions = function () {
    return SerialPortDeviceOptions;
}
//#endregion Device Option methods

// Serial Port methods - update serial port configuration
//#region Serial Port methods
// UpdateSerialPortSettingPort
//  This function updates the Serial Port settings for a port, but does not affect
//  the settings for a device on that port. Updates the serialport table in the database.
//  Will only delete the device information file (portX.json) if the device has been removed
//
//  - Parameters
//      - newPortSettings   - New port parameters
//      - callback          - Returns updated port parameters
//
//  - Updates local and persistent storage
//      - SerialPortsSettings[_port]    - Local serial port's settings
//      - SerialPortDeviceObjects[_port] - Local device object for port
//      - dB.serialport table[_port]    - Persistent serial port's settings in dB
//      - data/port<_port>.json         - Persistent device model for port *** deletes file if device removed from port ***
module.exports.UpdateSerialPortSettingPort = UpdateSerialPortSettingPort;
function UpdateSerialPortSettingPort(newPortSettings, callback) {
    //---------------------------------------------------------------------
    //  Updates of the port have the following cases:
    //      Case #1:    new=NONE        old=NONE
    //      Case #2:    new=NONE        old=Device
    //      Case #3:    new=Device      old=NONE
    //      Case #4:    new=Device      old=Device (new!=old)
    //      Case #5:    new=Device      old=Device (new==old)

    // Set the new device and track previous device
    var _port = newPortSettings.Address;
    var _device = newPortSettings.Device;
    Logger.info("serialports.UpdateSerialPortSettingsPort: Port[" + _port + "]=" + SerialPortSettings[_port].Device + " -> " + _device);

    // Case #1 & #2: New Device=SERIAL_PORT_NO_DEVICE - we're clearing this Port
    //#region Case #1 & #2: New Device=SERIAL_PORT_NO_DEVICE - we're clearing this Port
    if (_device === SERIAL_PORT_NO_DEVICE) {

        // Case #2a: If there was a different device on this newPortSettings, stop and remove it
        //#region Case #2a: If there was a different device on this newPortSettings, stop and remove it
        if (SerialPortSettings[_port].Device !== SERIAL_PORT_NO_DEVICE) {
            // Stop processing in a device was previously on this port
            SerialPortDeviceObjects[_port].Enable(false);// Stop the previous device
            SerialPortDeviceObjects[_port].SendEvents(false);// Stop sending events from the device
            Logger.info("serialports.UpdateSerialPortSettingsPort: Port[" + _port + "]=" + SerialPortSettings[_port].Device + " was stopped!");
        }
        //#endregion Case #2a: If there was a different device on this newPortSettings, stop and remove it

        // Case #1 & #2b: De-instantiate previous device and set port to empty
        //#region Case #1 & #2b: De-instantiate previous device and set port to empty
        SerialPortSettings[_port].Name = "Port " + (Number(_port) + 1);
        SerialPortSettings[_port].Enabled = 'FALSE';
        SerialPortSettings[_port].SendEvents = 'FALSE';
        SerialPortSettings[_port].Device = SERIAL_PORT_NO_DEVICE;
        SerialPortSettings[_port].TimeChanged = moment();// Device has changed

        SerialPortDeviceObjects[_port] = null;

        // Delete the persistent device model file
        //#region Delete the persistent device model file
        if (fs.existsSync(SERIAL_PORT_DEVICE_FILE + _port + ".json")) {
            // Delete previous persistence settings and log files for NULL devices
            fs.unlinkSync(SERIAL_PORT_DEVICE_FILE + _port + ".json");

            // Do a sync to ensure that the files are taken care of
            cp.exec("sync", function (error, stdout, stderr) {
                if (error) {
                    Logger.error("serialports.UpdateSerialPortSettingsPort: Removed and sync'ed " + SERIAL_PORT_DEVICE_FILE + _port + ".json error=" + error);
                }
                Logger.info("serialports.UpdateSerialPortSettingsPort: Removed and sync'ed " + SERIAL_PORT_DEVICE_FILE + _port + ".json");
            });
        }
        //#endregion Delete the persistent device model file

        Logger.info("serialports.UpdateSerialPortSettingsPort: Port[" + _port + "]=" + SerialPortSettings[_port].Device + " was removed!");
        //#endregion Case #1 & #2b: De-instantiate previous device and set port to empty
    }
    //#endregion Case #1 & #2: New Device=SERIAL_PORT_NO_DEVICE - we're clearing this Port
    else
    // Case #3, #4 & #5: New Device=!SERIAL_PORT_NO_DEVICE (new device, replacing device or modifying existing device)
    //#region Case #3, #4 & #5: New Device=!SERIAL_PORT_NO_DEVICE (new device, replacing device or modifying existing device)
    {
        // Case#4: New Device != Previous Device - first stop and remove old device, then add new device
        //#region Case#4: New Device != Previous Device - first stop and remove old device, then add new device
        if (_device != SerialPortSettings[_port].Device) {

            // Case #4a: If there was a different device on this newPortSettings, stop and remove it
            //#region Case #4a: If there was a different device on this newPortSettings, stop and remove it
            if (SerialPortSettings[_port].Device !== SERIAL_PORT_NO_DEVICE) {
                // Stop processing in a device was previously on this port
                SerialPortDeviceObjects[_port].Enable(false);// Stop the previous device
                SerialPortDeviceObjects[_port].SendEvents(false);// Stop sending events from the device
                Logger.info("serialports.UpdateSerialPortSettingsPort: Port[" + _port + "]=" + SerialPortSettings[_port].Device + " was stopped!");
            }
            //#endregion Case #4a: If there was a different device on this newPortSettings, stop and remove it

            // Case #3 & #4b: Add the new device object instantiation to SerialPortSettings
            //#region Case #3 & #4b: Add the new device object instantiation to SerialPortSettings
            require("./devices/" + _device + ".js")(_port, SERIAL_PORT_LOAD_DEFAULT_MODEL, Logger, SocketIo, SmtpManager, NetSnmp, function (deviceObject) {
                // Update settings
                SerialPortSettings[_port].Name = newPortSettings.Name;
                SerialPortSettings[_port].Enabled = newPortSettings.Enabled;
                SerialPortSettings[_port].SendEvents = newPortSettings.SendEvents;
                SerialPortSettings[_port].Device = newPortSettings.Device;
                SerialPortSettings[_port].TimeChanged = moment();// Device has changed

                SerialPortDeviceObjects[_port] = deviceObject;// Unique device object
                var _devicemodel = SerialPortDeviceObjects[_port].GetDeviceModel();// Default device model

                // Enable processing/eventing
                SerialPortDeviceObjects[_port].Enable(SerialPortSettings[_port].Enabled);// Start or stop the device
                SerialPortDeviceObjects[_port].SendEvents(SerialPortSettings[_port].SendEvents);// SendEvents or not from the device
                SerialPortDeviceObjects[_port].SetSnmpHostSessionsArray(SnmpHostSessionsArray, function () { });// Update the SNMP Sessions Array
                Logger.info("serialports.UpdateSerialPortSettingsPort: Port[" + _port + "]=" + SerialPortSettings[_port].Device + " (uuid=" + _devicemodel.uuid + ") was instantiated!");
            });
            //#endregion Case #3 & #4b: Add the new device object instantiation to SerialPortSettings
        }
        //#endregion Case#4: New Device != Previous Device - first stop and remove old device, then add new device

        // Case #5: New Device == Previous Device - use existing device instantiation, modify parameters
        //#region Case #5: New Device == Previous Device - use existing device instantiation, modify parameters
        else {
            // Update settings
            SerialPortSettings[_port].Name = newPortSettings.Name;
            SerialPortSettings[_port].Device = newPortSettings.Device;
            SerialPortSettings[_port].Enabled = newPortSettings.Enabled;
            SerialPortSettings[_port].SendEvents = newPortSettings.SendEvents;
            SerialPortSettings[_port].TimeChanged = moment();// Device was modified

            // DeviceObject doesn't change!!!
            var _devicemodel = SerialPortDeviceObjects[_port].GetDeviceModel();

            // Enable processing/eventing
            SerialPortDeviceObjects[_port].Enable(SerialPortSettings[_port].Enabled);// Start or stop the device
            SerialPortDeviceObjects[_port].SendEvents(SerialPortSettings[_port].SendEvents);// SendEvents or not from the device
            SerialPortDeviceObjects[_port].SetSnmpHostSessionsArray(SnmpHostSessionsArray, function () { });// Update the SNMP Sessions Array
            Logger.info("serialports.UpdateSerialPortSettingsPort: Port[" + _port + "]=" + SerialPortSettings[_port].Device + "(" + _devicemodel.uuid + ") was instantiated!");
        }
        //#endregion Case #5: New Device == Previous Device - use existing device instantiation, modify parameters
    }
    //#endregion Case #3, #4 & #5: New Device=!SERIAL_PORT_NO_DEVICE (new device, replacing device or modifying existing device)

    // Update ports.json to ensure persistence of the newPortSettings parameters
    //#region Update database (port information) and portX.json (device information for a port) to ensure persistence of the newPortSettings parameters
    //#region Update the database (port information)
    SerialPortSettings[_port].TimeChanged = moment();// Record the create time as the last time the SerialPort was changed
    var update = "UPDATE SerialPorts SET Name='" + SerialPortSettings[_port].Name + "', Enabled='" + SerialPortSettings[_port].Enabled + "', SendEvents='" + SerialPortSettings[_port].SendEvents + "', Device='" + SerialPortSettings[_port].Device + "', TimeChanged=" + SerialPortSettings[_port].TimeChanged + " WHERE address='" + SerialPortSettings[_port].Address + "'";
    DbManager.run(update, function (error) {
        if (error !== null) {
            Logger.error("serialports.UpdateSerialPortSettingsPort: Port[" + _port + "]=" + SerialPortSettings[_port].Device + " Error updating the database=" + error);
        }
    });
    //#endregion Update the database (port information)

    // Update the clients with changes
    //#region Update the clients with changes
    SocketIo.sockets.emit("Port", { "Port": SerialPortSettings[_port] });
    callback(SerialPortSettings[_port]);
    //#endregion Update the clients with changes

    //#endregion Update database (port information) and portX.json (device information for a port) to ensure persistence of the newPortSettings parameters
}

//  - Reset port
//      - port                          - Port to reset
module.exports.ResetSerialPortSettingPort = ResetSerialPortSettingPort;
function ResetSerialPortSettingPort(port, callback) {
    // Only make a call to reset the device, if there is a device allocated
    if (SerialPortDeviceObjects[port] != null) {
        SerialPortDeviceObjects[port].Reset(function (status) {
            Logger.debug("serialports.ResetSerialPortSettingPort: Reset device - status=" + status);
            callback(status)
        })
    }
}

//  - Zero port
//      - port                          - Port to zero
module.exports.ZeroSerialPortSettingPort = ZeroSerialPortSettingPort;
function ZeroSerialPortSettingPort(port, callback) {
    // Only make a call to reset the device, if there is a device allocated
    if (SerialPortDeviceObjects[port] != null) {
        SerialPortDeviceObjects[port].Zero(function (status) {
            Logger.debug("serialports.ZeroSerialPortSettingPort: Zero device - status=" + status);
            callback(status)
        })
    }
}

// Update the SNMP Host array
module.exports.SetSnmpHostSessionsArray = SetSnmpHostSessionsArray;
function SetSnmpHostSessionsArray(snmpHostSessions,callback) {
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array

    // Copy the SNMP session to all of the active serial ports
    for (var i = 0; i < SerialPortDeviceObjects.length; i++) {
        if (SerialPortDeviceObjects[i] != null) {
            SerialPortDeviceObjects[i].SetSnmpHostSessionsArray(SnmpHostSessionsArray, function () { })
        }
    }
    Logger.info("serialports.UpdateSerialPortSettingsPort: Initialized SNMP Host Sessions");

    callback();
}

// Update Qualifier Values
//  This function updates the qualifiers for a serial port
module.exports.UpdateSerialPortQualifiers = UpdateSerialPortQualifiers
function UpdateSerialPortQualifiers(digitalInputs) {
    for (var i = 0; i < SerialPortDeviceObjects.length; i++) {
        if (SerialPortDeviceObjects[i] != null) {
            SerialPortDeviceObjects[i].UpdateQualifierValues(digitalInputs);
        }
    }
}
//#endregion Serial Port methods

// Device Model methods - get/update and manage devices
//#region Device Model methods

// GetSerialPortDeviceModel
//  Gets the current device specific settings
module.exports.GetSerialPortDeviceModel = GetSerialPortDeviceModel;
function GetSerialPortDeviceModel(port) {

    // Get current device model
    return SerialPortDeviceObjects[port].GetDeviceModel();
}

// UpdateSerialPortDeviceModel
//  This function updates device specific settings, for example:
//      - Hold Times
//      - Qualifiers
//      - Formulas
//
//  - Parameters
//      - _port             - Serial port being modified
//      - newDeviceModel    - New port parameters
//      - callback          - Returns updated port parameters
module.exports.UpdateSerialPortDeviceModel = UpdateSerialPortDeviceModel
function UpdateSerialPortDeviceModel(_port, newDeviceModel, callback) {

    // Update UpdateSerialPortDeviceModel
    //#region Update UpdateSerialPortDeviceModel
    SerialPortDeviceObjects[_port].UpdateDeviceModel(_port, newDeviceModel, function (devicemodel) {
        callback(devicemodel);
    });
    //#endregion Update UpdateSerialPortDeviceModel
}
//#endregion Device Model methods

//#endregion Methods

